# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import deque
from functools import partial
import logging
from os import path
import re
from typing import Dict, List, Optional
from urllib.parse import unquote, urlparse

from _pytest.fixtures import FixtureRequest
import pytest
import requests
from requests.adapters import BaseAdapter
from requests.structures import CaseInsensitiveDict
from requests.utils import get_encoding_from_headers

logger = logging.getLogger(__name__)


# Check get_local_factory function
# Maximum number of iteration checks to generate requests responses
MAX_VISIT_FILES = 10


def get_response_cb(
    request: requests.Request,
    context,
    datadir,
    ignore_urls: List[str] = [],
    visits: Optional[Dict] = None,
):
    """Mount point callback to fetch on disk the request's content. The request
    urls provided are url decoded first to resolve the associated file on disk.

    This is meant to be used as 'body' argument of the requests_mock.get()
    method.

    It will look for files on the local filesystem based on the requested URL,
    using the following rules:

    - files are searched in the datadir/<hostname> directory

    - the local file name is the path part of the URL with path hierarchy
      markers (aka '/') replaced by '_'

    Eg. if you use the requests_mock fixture in your test file as:

        requests_mock.get('https?://nowhere.com', body=get_response_cb)
        # or even
        requests_mock.get(re.compile('https?://'), body=get_response_cb)

    then a call requests.get like:

        requests.get('https://nowhere.com/path/to/resource?a=b&c=d')

    will look the content of the response in:

        datadir/https_nowhere.com/path_to_resource,a=b,c=d

    or a call requests.get like:

        requests.get('http://nowhere.com/path/to/resource?a=b&c=d')

    will look the content of the response in:

        datadir/http_nowhere.com/path_to_resource,a=b,c=d

    Args:
        request: Object requests
        context (requests.Context): Object holding response metadata
            information (status_code, headers, etc...)
        datadir: Data files path
        ignore_urls: urls whose status response should be 404 even if the local
            file exists
        visits: Dict of url, number of visits. If None, disable multi visit
            support (default)

    Returns:
        Optional[FileDescriptor] on disk file to read from the test context

    """
    logger.debug("get_response_cb(%s, %s)", request, context)
    logger.debug("url: %s", request.url)
    logger.debug("ignore_urls: %s", ignore_urls)
    unquoted_url = unquote(request.url)
    if unquoted_url in ignore_urls:
        context.status_code = 404
        return None
    url = urlparse(unquoted_url)
    # http://pypi.org ~> http_pypi.org
    # https://files.pythonhosted.org ~> https_files.pythonhosted.org
    dirname = "%s_%s" % (url.scheme, url.hostname)
    # url.path: pypi/<project>/json -> local file: pypi_<project>_json
    filename = url.path[1:]
    if filename.endswith("/"):
        filename = filename[:-1]
    filename = filename.replace("/", "_")
    if url.query:
        filename += "," + url.query.replace("&", ",")

    filepath = path.join(datadir, dirname, filename)
    if visits is not None:
        visit = visits.get(url, 0)
        visits[url] = visit + 1
        if visit:
            filepath = filepath + "_visit%s" % visit

    if not path.isfile(filepath):
        logger.debug("not found filepath: %s", filepath)
        context.status_code = 404
        return None
    fd = open(filepath, "rb")
    context.headers["content-length"] = str(path.getsize(filepath))
    return fd


@pytest.fixture
def datadir(request: FixtureRequest) -> str:
    """By default, returns the test directory's data directory.

    This can be overridden on a per file tree basis. Add an override
    definition in the local conftest, for example::

        import pytest

        from os import path

        @pytest.fixture
        def datadir():
            return path.join(path.abspath(path.dirname(__file__)), 'resources')


    """
    return path.join(path.dirname(str(request.fspath)), "data")


def requests_mock_datadir_factory(
    ignore_urls: List[str] = [], has_multi_visit: bool = False
):
    """This factory generates fixtures which allow to look for files on the
    local filesystem based on the requested URL, using the following rules:

    - files are searched in the data/<hostname> directory

    - the local file name is the path part of the URL with path hierarchy
      markers (aka '/') replaced by '_'

    Multiple implementations are possible, for example:

    ``requests_mock_datadir_factory([])``

        This computes the file name from the query and always returns the same
        result.

    ``requests_mock_datadir_factory(has_multi_visit=True)``

        This computes the file name from the query and returns the content of
        the filename the first time, the next call returning the content of
        files suffixed with _visit1 and so on and so forth. If the file is not
        found, returns a 404.

    ``requests_mock_datadir_factory(ignore_urls=['url1', 'url2'])``

        This will ignore any files corresponding to url1 and url2, always
        returning 404.

    Args:
        ignore_urls: List of urls to always returns 404 (whether file
            exists or not)
        has_multi_visit: Activate or not the multiple visits behavior

    """

    @pytest.fixture
    def requests_mock_datadir(requests_mock, datadir):
        if not has_multi_visit:
            cb = partial(get_response_cb, ignore_urls=ignore_urls, datadir=datadir)
            requests_mock.get(re.compile("https?://"), body=cb)
        else:
            visits = {}
            requests_mock.get(
                re.compile("https?://"),
                body=partial(
                    get_response_cb,
                    ignore_urls=ignore_urls,
                    visits=visits,
                    datadir=datadir,
                ),
            )

        return requests_mock

    return requests_mock_datadir


# Default `requests_mock_datadir` implementation
requests_mock_datadir = requests_mock_datadir_factory()
"""
Instance of :py:func:`requests_mock_datadir_factory`,
with the default arguments.
"""

# Implementation for multiple visits behavior:
# - first time, it checks for a file named `filename`
# - second time, it checks for a file named `filename`_visit1
# etc...
requests_mock_datadir_visits = requests_mock_datadir_factory(has_multi_visit=True)
"""
Instance of :py:func:`requests_mock_datadir_factory`,
with the default arguments, but `has_multi_visit=True`.
"""


@pytest.fixture
def swh_rpc_client(swh_rpc_client_class, swh_rpc_adapter):
    """This fixture generates an RPCClient instance that uses the class generated
    by the rpc_client_class fixture as backend.

    Since it uses the swh_rpc_adapter, HTTP queries will be intercepted and
    routed directly to the current Flask app (as provided by the `app`
    fixture).

    So this stack of fixtures allows to test the RPCClient -> RPCServerApp
    communication path using a real RPCClient instance and a real Flask
    (RPCServerApp) app instance.

    To use this fixture:

    - ensure an `app` fixture exists and generate a Flask application,
    - implement an `swh_rpc_client_class` fixtures that returns the
      RPCClient-based class to use as client side for the tests,
    - implement your tests using this `swh_rpc_client` fixture.

    See swh/core/api/tests/test_rpc_client_server.py for an example of usage.
    """
    url = "mock://example.com"
    cli = swh_rpc_client_class(url=url)
    # we need to clear the list of existing adapters here so we ensure we
    # have one and only one adapter which is then used for all the requests.
    cli.session.adapters.clear()
    cli.session.mount("mock://", swh_rpc_adapter)
    return cli


@pytest.fixture
def swh_rpc_adapter(app):
    """Fixture that generates a requests.Adapter instance that
    can be used to test client/servers code based on swh.core.api classes.

    See swh/core/api/tests/test_rpc_client_server.py for an example of usage.

    """
    with app.test_client() as client:
        yield RPCTestAdapter(client)


class RPCTestAdapter(BaseAdapter):
    def __init__(self, client):
        self._client = client

    def build_response(self, req, resp):
        response = requests.Response()

        # Fallback to None if there's no status_code, for whatever reason.
        response.status_code = resp.status_code

        # Make headers case-insensitive.
        response.headers = CaseInsensitiveDict(getattr(resp, "headers", {}))

        # Set encoding.
        response.encoding = get_encoding_from_headers(response.headers)
        response.raw = resp
        response.reason = response.raw.status

        if isinstance(req.url, bytes):
            response.url = req.url.decode("utf-8")
        else:
            response.url = req.url

        # Give the Response some context.
        response.request = req
        response.connection = self
        response._content = resp.data

        return response

    def send(self, request, **kw):
        """
        Overrides ``requests.adapters.BaseAdapter.send``
        """
        resp = self._client.open(
            request.url,
            method=request.method,
            headers=request.headers.items(),
            data=request.body,
        )
        return self.build_response(request, resp)


@pytest.fixture
def flask_app_client(app):
    with app.test_client() as client:
        yield client


# stolen from pytest-flask, required to have url_for() working within tests
# using flask_app_client fixture.
@pytest.fixture(autouse=True)
def _push_request_context(request: FixtureRequest):
    """During tests execution request context has been pushed, e.g. `url_for`,
    `session`, etc. can be used in tests as is::

        def test_app(app, client):
            assert client.get(url_for('myview')).status_code == 200

    """
    if "app" not in request.fixturenames:
        return
    app = request.getfixturevalue("app")
    ctx = app.test_request_context()
    ctx.push()

    def teardown():
        ctx.pop()

    request.addfinalizer(teardown)


class FakeSocket(object):
    """ A fake socket for testing. """

    def __init__(self):
        self.payloads = deque()

    def send(self, payload):
        assert type(payload) == bytes
        self.payloads.append(payload)

    def recv(self):
        try:
            return self.payloads.popleft().decode("utf-8")
        except IndexError:
            return None

    def close(self):
        pass

    def __repr__(self):
        return str(self.payloads)


@pytest.fixture
def statsd():
    """Simple fixture giving a Statsd instance suitable for tests

    The Statsd instance uses a FakeSocket as `.socket` attribute in which one
    can get the accumulated statsd messages in a deque in `.socket.payloads`.
    """

    from swh.core.statsd import Statsd

    statsd = Statsd()
    statsd._socket = FakeSocket()
    yield statsd



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
import datetime
import enum
import json
import logging
import os
import sys
import threading
from typing import Any, Callable, Iterable, Iterator, Mapping, Optional, Type, TypeVar

import psycopg2
import psycopg2.extras
import psycopg2.pool

logger = logging.getLogger(__name__)


psycopg2.extras.register_uuid()


def render_array(data) -> str:
    """Render the data as a postgresql array"""
    # From https://www.postgresql.org/docs/11/arrays.html#ARRAYS-IO
    # "The external text representation of an array value consists of items that are
    # interpreted according to the I/O conversion rules for the array's element type,
    # plus decoration that indicates the array structure. The decoration consists of
    # curly braces ({ and }) around the array value plus delimiter characters between
    # adjacent items. The delimiter character is usually a comma (,)"
    return "{%s}" % ",".join(render_array_element(e) for e in data)


def render_array_element(element) -> str:
    """Render an element from an array."""
    if element is None:
        # From https://www.postgresql.org/docs/11/arrays.html#ARRAYS-IO
        # "If the value written for an element is NULL (in any case variant), the
        # element is taken to be NULL."
        return "NULL"
    elif isinstance(element, (list, tuple)):
        # From https://www.postgresql.org/docs/11/arrays.html#ARRAYS-INPUT
        # "Each val is either a constant of the array element type, or a subarray."
        return render_array(element)
    else:
        # From https://www.postgresql.org/docs/11/arrays.html#ARRAYS-IO
        # "When writing an array value you can use double quotes around any individual
        # array element. [...] Empty strings and strings matching the word NULL must be
        # quoted, too. To put a double quote or backslash in a quoted array element
        # value, precede it with a backslash."
        ret = value_as_pg_text(element)
        return '"%s"' % ret.replace("\\", "\\\\").replace('"', '\\"')


def value_as_pg_text(data: Any) -> str:
    """Render the given data in the postgresql text format.

    NULL values are handled **outside** of this function (either by
    :func:`render_array_element`, or by :meth:`BaseDb.copy_to`.)
    """

    if data is None:
        raise ValueError("value_as_pg_text doesn't handle NULLs")

    if isinstance(data, bytes):
        return "\\x%s" % data.hex()
    elif isinstance(data, datetime.datetime):
        return data.isoformat()
    elif isinstance(data, dict):
        return json.dumps(data)
    elif isinstance(data, (list, tuple)):
        return render_array(data)
    elif isinstance(data, psycopg2.extras.Range):
        return "%s%s,%s%s" % (
            "[" if data.lower_inc else "(",
            "-infinity" if data.lower_inf else value_as_pg_text(data.lower),
            "infinity" if data.upper_inf else value_as_pg_text(data.upper),
            "]" if data.upper_inc else ")",
        )
    elif isinstance(data, enum.IntEnum):
        return str(int(data))
    else:
        return str(data)


def escape_copy_column(column: str) -> str:
    """Escape the text representation of a column for use by COPY."""
    # From https://www.postgresql.org/docs/11/sql-copy.html
    # File Formats > Text Format
    # "Backslash characters (\) can be used in the COPY data to quote data characters
    # that might otherwise be taken as row or column delimiters. In particular, the
    # following characters must be preceded by a backslash if they appear as part of a
    # column value: backslash itself, newline, carriage return, and the current
    # delimiter character."
    ret = (
        column.replace("\\", "\\\\")
        .replace("\n", "\\n")
        .replace("\r", "\\r")
        .replace("\t", "\\t")
    )

    return ret


def typecast_bytea(value, cur):
    if value is not None:
        data = psycopg2.BINARY(value, cur)
        return data.tobytes()


BaseDbType = TypeVar("BaseDbType", bound="BaseDb")


class BaseDb:
    """Base class for swh.*.*Db.

    cf. swh.storage.db.Db, swh.archiver.db.ArchiverDb

    """

    @staticmethod
    def adapt_conn(conn: psycopg2.extensions.connection):
        """Makes psycopg2 use 'bytes' to decode bytea instead of
        'memoryview', for this connection."""
        t_bytes = psycopg2.extensions.new_type((17,), "bytea", typecast_bytea)
        psycopg2.extensions.register_type(t_bytes, conn)

        t_bytes_array = psycopg2.extensions.new_array_type((1001,), "bytea[]", t_bytes)
        psycopg2.extensions.register_type(t_bytes_array, conn)

    @classmethod
    def connect(cls: Type[BaseDbType], *args, **kwargs) -> BaseDbType:
        """factory method to create a DB proxy

        Accepts all arguments of psycopg2.connect; only some specific
        possibilities are reported below.

        Args:
            connstring: libpq2 connection string

        """
        conn = psycopg2.connect(*args, **kwargs)
        return cls(conn)

    @classmethod
    def from_pool(
        cls: Type[BaseDbType], pool: psycopg2.pool.AbstractConnectionPool
    ) -> BaseDbType:
        conn = pool.getconn()
        return cls(conn, pool=pool)

    def __init__(
        self,
        conn: psycopg2.extensions.connection,
        pool: Optional[psycopg2.pool.AbstractConnectionPool] = None,
    ):
        """create a DB proxy

        Args:
            conn: psycopg2 connection to the SWH DB
            pool: psycopg2 pool of connections

        """
        self.adapt_conn(conn)
        self.conn = conn
        self.pool = pool

    def put_conn(self) -> None:
        if self.pool:
            self.pool.putconn(self.conn)

    def cursor(
        self, cur_arg: Optional[psycopg2.extensions.cursor] = None
    ) -> psycopg2.extensions.cursor:
        """get a cursor: from cur_arg if given, or a fresh one otherwise

        meant to avoid boilerplate if/then/else in methods that proxy stored
        procedures

        """
        if cur_arg is not None:
            return cur_arg
        else:
            return self.conn.cursor()

    _cursor = cursor  # for bw compat

    @contextmanager
    def transaction(self) -> Iterator[psycopg2.extensions.cursor]:
        """context manager to execute within a DB transaction

        Yields:
            a psycopg2 cursor

        """
        with self.conn.cursor() as cur:
            try:
                yield cur
                self.conn.commit()
            except Exception:
                if not self.conn.closed:
                    self.conn.rollback()
                raise

    def copy_to(
        self,
        items: Iterable[Mapping[str, Any]],
        tblname: str,
        columns: Iterable[str],
        cur: Optional[psycopg2.extensions.cursor] = None,
        item_cb: Optional[Callable[[Any], Any]] = None,
        default_values: Optional[Mapping[str, Any]] = None,
    ) -> None:
        """Run the COPY command to insert the `columns` of each element of `items` into
        `tblname`.

        Args:
            items: dictionaries of data to copy into `tblname`.
            tblname: name of the destination table.
            columns: columns of the destination table. Elements of `items` must have
              these set as keys.
            default_values: dictionary of default values to use when inserting entries
              in `tblname`.
            cur: a db cursor; if not given, a new cursor will be created.
            item_cb: optional callback, run on each element of `items`, when it is
              copied.

        """
        if default_values is None:
            default_values = {}

        read_file, write_file = os.pipe()
        exc_info = None

        def writer():
            nonlocal exc_info
            cursor = self.cursor(cur)
            with open(read_file, "r") as f:
                try:
                    cursor.copy_expert(
                        "COPY %s (%s) FROM STDIN" % (tblname, ", ".join(columns)), f
                    )
                except Exception:
                    # Tell the main thread about the exception
                    exc_info = sys.exc_info()

        write_thread = threading.Thread(target=writer)
        write_thread.start()

        try:
            with open(write_file, "w") as f:
                # From https://www.postgresql.org/docs/11/sql-copy.html
                # File Formats > Text Format
                # "When the text format is used, the data read or written is a text file
                # with one line per table row. Columns in a row are separated by the
                # delimiter character."
                # NULL
                # "The default is \N (backslash-N) in text format."
                # DELIMITER
                # "The default is a tab character in text format."
                for d in items:
                    if item_cb is not None:
                        item_cb(d)
                    line = []
                    for k in columns:
                        value = d.get(k, default_values.get(k))
                        try:
                            if value is None:
                                line.append("\\N")
                            else:
                                line.append(escape_copy_column(value_as_pg_text(value)))
                        except Exception as e:
                            logger.error(
                                "Could not escape value `%r` for column `%s`:"
                                "Received exception: `%s`",
                                value,
                                k,
                                e,
                            )
                            raise e from None
                    f.write("\t".join(line))
                    f.write("\n")

        finally:
            # No problem bubbling up exceptions, but we still need to make sure
            # we finish copying, even though we're probably going to cancel the
            # transaction.
            write_thread.join()
            if exc_info:
                # postgresql returned an error, let's raise it.
                raise exc_info[1].with_traceback(exc_info[2])

    def mktemp(self, tblname: str, cur: Optional[psycopg2.extensions.cursor] = None):
        self.cursor(cur).execute("SELECT swh_mktemp(%s)", (tblname,))



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import glob
from importlib import import_module
import logging
import subprocess
from typing import List, Optional, Set, Union

from _pytest.fixtures import FixtureRequest
import psycopg2
import pytest
from pytest_postgresql.janitor import DatabaseJanitor

from swh.core.utils import numfile_sortkey as sortkey

# to keep mypy happy regardless pytest-postgresql version
try:
    _pytest_pgsql_get_config_module = import_module("pytest_postgresql.config")
except ImportError:
    # pytest_postgresql < 3.0.0
    _pytest_pgsql_get_config_module = import_module("pytest_postgresql.factories")

_pytest_postgresql_get_config = getattr(_pytest_pgsql_get_config_module, "get_config")


logger = logging.getLogger(__name__)


class SWHDatabaseJanitor(DatabaseJanitor):
    """SWH database janitor implementation with a a different setup/teardown policy than
    than the stock one. Instead of dropping, creating and initializing the database for
    each test, it creates and initializes the db once, then truncates the tables (and
    sequences) in between tests.

    This is needed to have acceptable test performances.

    """

    def __init__(
        self,
        user: str,
        host: str,
        port: str,
        dbname: str,
        version: Union[str, float],
        dump_files: Union[None, str, List[str]] = None,
        no_truncate_tables: Set[str] = set(),
    ) -> None:
        super().__init__(user, host, port, dbname, version)
        if not hasattr(self, "dbname") and hasattr(self, "db_name"):
            # pytest_postgresql < 3.0.0
            self.dbname = getattr(self, "db_name")
        if dump_files is None:
            self.dump_files = []
        elif isinstance(dump_files, str):
            self.dump_files = sorted(glob.glob(dump_files), key=sortkey)
        else:
            self.dump_files = dump_files
        # do no truncate the following tables
        self.no_truncate_tables = set(no_truncate_tables)

    def db_setup(self):
        conninfo = (
            f"host={self.host} user={self.user} port={self.port} dbname={self.dbname}"
        )

        for fname in self.dump_files:
            subprocess.check_call(
                [
                    "psql",
                    "--quiet",
                    "--no-psqlrc",
                    "-v",
                    "ON_ERROR_STOP=1",
                    "-d",
                    conninfo,
                    "-f",
                    fname,
                ]
            )

    def db_reset(self):
        """Truncate tables (all but self.no_truncate_tables set) and sequences

        """
        with psycopg2.connect(
            dbname=self.dbname, user=self.user, host=self.host, port=self.port,
        ) as cnx:
            with cnx.cursor() as cur:
                cur.execute(
                    "SELECT table_name FROM information_schema.tables "
                    "WHERE table_schema = %s",
                    ("public",),
                )
                all_tables = set(table for (table,) in cur.fetchall())
                tables_to_truncate = all_tables - self.no_truncate_tables

                for table in tables_to_truncate:
                    cur.execute("TRUNCATE TABLE %s CASCADE" % table)

                cur.execute(
                    "SELECT sequence_name FROM information_schema.sequences "
                    "WHERE sequence_schema = %s",
                    ("public",),
                )
                seqs = set(seq for (seq,) in cur.fetchall())
                for seq in seqs:
                    cur.execute("ALTER SEQUENCE %s RESTART;" % seq)
            cnx.commit()

    def init(self):
        """Initialize db. Create the db if it does not exist. Reset it if it exists."""
        with self.cursor() as cur:
            cur.execute(
                "SELECT COUNT(1) FROM pg_database WHERE datname=%s;", (self.dbname,)
            )
            db_exists = cur.fetchone()[0] == 1
            if db_exists:
                cur.execute(
                    "UPDATE pg_database SET datallowconn=true WHERE datname = %s;",
                    (self.dbname,),
                )
                self.db_reset()
                return

        # initialize the inexistent db
        with self.cursor() as cur:
            cur.execute('CREATE DATABASE "{}";'.format(self.dbname))
        self.db_setup()

    def drop(self):
        """The original DatabaseJanitor implementation prevents new connections from happening,
           destroys current opened connections and finally drops the database.

           We actually do not want to drop the db so we instead do nothing and resets
           (truncate most tables and sequences) the db instead, in order to have some
           acceptable performance.

        """
        pass


# the postgres_fact factory fixture below is mostly a copy of the code
# from pytest-postgresql. We need a custom version here to be able to
# specify our version of the DBJanitor we use.
def postgresql_fact(
    process_fixture_name: str,
    dbname: Optional[str] = None,
    dump_files: Union[str, List[str]] = "",
    no_truncate_tables: Set[str] = {"dbversion"},
):
    @pytest.fixture
    def postgresql_factory(request: FixtureRequest):
        """Fixture factory for PostgreSQL.

        :param FixtureRequest request: fixture request object
        :rtype: psycopg2.connection
        :returns: postgresql client
        """
        config = _pytest_postgresql_get_config(request)
        proc_fixture = request.getfixturevalue(process_fixture_name)

        pg_host = proc_fixture.host
        pg_port = proc_fixture.port
        pg_user = proc_fixture.user
        pg_options = proc_fixture.options
        pg_db = dbname or config["dbname"]
        with SWHDatabaseJanitor(
            pg_user,
            pg_host,
            pg_port,
            pg_db,
            proc_fixture.version,
            dump_files=dump_files,
            no_truncate_tables=no_truncate_tables,
        ):
            connection = psycopg2.connect(
                dbname=pg_db,
                user=pg_user,
                host=pg_host,
                port=pg_port,
                options=pg_options,
            )
            yield connection
            connection.close()

    return postgresql_factory



# Copyright (C) 2015  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import logging
from typing import Any, Dict, Generator, List, Tuple

from systemd.journal import JournalHandler as _JournalHandler
from systemd.journal import send

try:
    from celery import current_task
except ImportError:
    current_task = None


EXTRA_LOGDATA_PREFIX = "swh_"
LOGGED_TASK_KWARGS = ("url", "instance")


def db_level_of_py_level(lvl):
    """convert a log level of the logging module to a log level suitable for the
    logging Postgres DB

    """
    return logging.getLevelName(lvl).lower()


def get_extra_data(record: logging.LogRecord) -> Dict[str, Any]:
    """Get the extra data to send to the log handler from the logging record.

    This gets the following data:
      - all fields in the record data starting with `EXTRA_LOGDATA_PREFIX`
      - arguments to the logging call (which can either be a tuple, or a dict
        if the arguments were named)
      - if this is called within a celery task, the following data:
        - the (uu)id of the task
        - the name of the task
        - any task keyword arguments named for values in `LOGGED_TASK_KWARGS`
    """
    log_data = record.__dict__

    extra_data = {
        k[len(EXTRA_LOGDATA_PREFIX) :]: v
        for k, v in log_data.items()
        if k.startswith(EXTRA_LOGDATA_PREFIX)
    }

    args = log_data.get("args")
    if args:
        extra_data["logging_args"] = args

    # Retrieve Celery task info
    if current_task and current_task.request:
        extra_data["task"] = {
            "id": current_task.request.id,
            "name": current_task.name,
        }

        for task_arg in LOGGED_TASK_KWARGS:
            if task_arg in current_task.request.kwargs:
                try:
                    value = stringify(current_task.request.kwargs[task_arg])
                except Exception:
                    continue

                extra_data["task"][f"kwarg_{task_arg}"] = value

    return extra_data


def flatten(data: Any, separator: str = "_") -> Generator[Tuple[str, Any], None, None]:
    """Flatten the data dictionary into a flat structure"""

    def inner_flatten(
        data: Any, prefix: List[str]
    ) -> Generator[Tuple[List[str], Any], None, None]:
        if isinstance(data, dict):
            if all(isinstance(key, str) for key in data):
                for key, value in data.items():
                    yield from inner_flatten(value, prefix + [key])
            else:
                yield prefix, str(data)
        elif isinstance(data, (list, tuple)):
            for key, value in enumerate(data):
                yield from inner_flatten(value, prefix + [str(key)])
        else:
            yield prefix, data

    for path, value in inner_flatten(data, []):
        yield separator.join(path), value


def stringify(value: Any) -> str:
    """Convert value to string"""
    if isinstance(value, datetime.datetime):
        return value.isoformat()

    return str(value)


class JournalHandler(_JournalHandler):
    def emit(self, record):
        """Write `record` as a journal event.

        MESSAGE is taken from the message provided by the user, and PRIORITY,
        LOGGER, THREAD_NAME, CODE_{FILE,LINE,FUNC} fields are appended
        automatically. In addition, record.MESSAGE_ID will be used if present.

        This also records all the extra data fetched by `get_extra_data`.
        """
        try:
            extra_data = flatten(get_extra_data(record))
            extra_data = {
                (EXTRA_LOGDATA_PREFIX + key).upper(): stringify(value)
                for key, value in extra_data
            }
            msg = self.format(record)
            pri = self.mapPriority(record.levelno)
            send(
                msg,
                PRIORITY=format(pri),
                LOGGER=record.name,
                THREAD_NAME=record.threadName,
                CODE_FILE=record.pathname,
                CODE_LINE=record.lineno,
                CODE_FUNC=record.funcName,
                **extra_data,
            )
        except Exception:
            self.handleError(record)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re

import pytest
from requests.exceptions import ConnectionError

from swh.core.api import APIError, RemoteException, RPCClient, remote_api_endpoint
from swh.core.api.serializers import exception_to_dict, msgpack_dumps

from .test_serializers import ExtraType, extra_decoders, extra_encoders


class ReraiseException(Exception):
    pass


@pytest.fixture
def rpc_client(requests_mock):
    class TestStorage:
        @remote_api_endpoint("test_endpoint_url")
        def test_endpoint(self, test_data, db=None, cur=None):
            ...

        @remote_api_endpoint("path/to/endpoint")
        def something(self, data, db=None, cur=None):
            ...

        @remote_api_endpoint("serializer_test")
        def serializer_test(self, data, db=None, cur=None):
            ...

        @remote_api_endpoint("overridden/endpoint")
        def overridden_method(self, data):
            return "foo"

    class Testclient(RPCClient):
        backend_class = TestStorage
        extra_type_encoders = extra_encoders
        extra_type_decoders = extra_decoders
        reraise_exceptions = [ReraiseException]

        def overridden_method(self, data):
            return "bar"

    def callback(request, context):
        assert request.headers["Content-Type"] == "application/x-msgpack"
        context.headers["Content-Type"] = "application/x-msgpack"
        if request.path == "/test_endpoint_url":
            context.content = b"\xa3egg"
        elif request.path == "/path/to/endpoint":
            context.content = b"\xa4spam"
        elif request.path == "/serializer_test":
            context.content = (
                b"\x82\xc4\x07swhtype\xa9extratype"
                b"\xc4\x01d\x92\x81\xa4spam\xa3egg\xa3qux"
            )
        else:
            assert False
        return context.content

    requests_mock.post(re.compile("mock://example.com/"), content=callback)

    return Testclient(url="mock://example.com")


def test_client(rpc_client):

    assert hasattr(rpc_client, "test_endpoint")
    assert hasattr(rpc_client, "something")

    res = rpc_client.test_endpoint("spam")
    assert res == "egg"
    res = rpc_client.test_endpoint(test_data="spam")
    assert res == "egg"

    res = rpc_client.something("whatever")
    assert res == "spam"
    res = rpc_client.something(data="whatever")
    assert res == "spam"


def test_client_extra_serializers(rpc_client):
    res = rpc_client.serializer_test(["foo", ExtraType("bar", b"baz")])
    assert res == ExtraType({"spam": "egg"}, "qux")


def test_client_overridden_method(rpc_client):
    res = rpc_client.overridden_method("foo")
    assert res == "bar"


def test_client_connexion_error(rpc_client, requests_mock):
    """
    ConnectionError should be wrapped and raised as an APIError.
    """
    error_message = "unreachable host"
    requests_mock.post(
        re.compile("mock://example.com/connection_error"),
        exc=ConnectionError(error_message),
    )

    with pytest.raises(APIError) as exc_info:
        rpc_client.post("connection_error", data={})

    assert type(exc_info.value.args[0]) == ConnectionError
    assert str(exc_info.value.args[0]) == error_message


def _exception_response(exception, status_code, old_exception_schema=False):
    def callback(request, context):
        assert request.headers["Content-Type"] == "application/x-msgpack"
        context.headers["Content-Type"] = "application/x-msgpack"
        exc_dict = exception_to_dict(exception)
        if old_exception_schema:
            exc_dict = {"exception": exc_dict}
        context.content = msgpack_dumps(exc_dict)
        context.status_code = status_code
        return context.content

    return callback


@pytest.mark.parametrize("old_exception_schema", [False, True])
def test_client_reraise_exception(rpc_client, requests_mock, old_exception_schema):
    """
    Exception caught server-side and whitelisted will be raised again client-side.
    """
    error_message = "something went wrong"
    endpoint = "reraise_exception"

    requests_mock.post(
        re.compile(f"mock://example.com/{endpoint}"),
        content=_exception_response(
            exception=ReraiseException(error_message),
            status_code=400,
            old_exception_schema=old_exception_schema,
        ),
    )

    with pytest.raises(ReraiseException) as exc_info:
        rpc_client.post(endpoint, data={})

    assert str(exc_info.value) == error_message


@pytest.mark.parametrize(
    "status_code, old_exception_schema",
    [(400, False), (500, False), (400, True), (500, True),],
)
def test_client_raise_remote_exception(
    rpc_client, requests_mock, status_code, old_exception_schema
):
    """
    Exception caught server-side and not whitelisted will be wrapped and raised
    as a RemoteException client-side.
    """
    error_message = "something went wrong"
    endpoint = "raise_remote_exception"

    requests_mock.post(
        re.compile(f"mock://example.com/{endpoint}"),
        content=_exception_response(
            exception=Exception(error_message),
            status_code=status_code,
            old_exception_schema=old_exception_schema,
        ),
    )

    with pytest.raises(RemoteException) as exc_info:
        rpc_client.post(endpoint, data={})

    assert str(exc_info.value.args[0]["type"]) == "Exception"
    assert str(exc_info.value.args[0]["message"]) == error_message



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.core.api import remote_api_endpoint
from swh.core.api.asynchronous import RPCServerApp
from swh.core.api.serializers import json_dumps, msgpack_dumps

from .test_serializers import ExtraType, extra_decoders, extra_encoders


class MyRPCServerApp(RPCServerApp):
    extra_type_encoders = extra_encoders
    extra_type_decoders = extra_decoders


class BackendStorageTest:
    """Backend Storage to use as backend class of the rpc server (test only)"""

    @remote_api_endpoint("test_endpoint_url", method="GET")
    def test_endpoint(self, test_data, db=None, cur=None):
        assert test_data == "spam"
        return "egg"

    @remote_api_endpoint("path/to/identity")
    def identity(self, data, db=None, cur=None):
        return data

    @remote_api_endpoint("serializer_test")
    def serializer_test(self, data, db=None, cur=None):
        assert data == ["foo", ExtraType("bar", b"baz")]
        return ExtraType({"spam": "egg"}, "qux")


@pytest.fixture
def async_app():
    return MyRPCServerApp("testapp", backend_class=BackendStorageTest)


def test_api_async_rpc_server_app_ok(async_app):
    assert isinstance(async_app, MyRPCServerApp)

    actual_rpc_server2 = MyRPCServerApp(
        "app2", backend_class=BackendStorageTest, backend_factory=BackendStorageTest
    )
    assert isinstance(actual_rpc_server2, MyRPCServerApp)

    actual_rpc_server3 = MyRPCServerApp("app3")
    assert isinstance(actual_rpc_server3, MyRPCServerApp)


def test_api_async_rpc_server_app_misconfigured():
    expected_error = "backend_factory should only be provided if backend_class is"
    with pytest.raises(ValueError, match=expected_error):
        MyRPCServerApp("failed-app", backend_factory="something-to-make-it-raise")


@pytest.fixture
def cli(loop, aiohttp_client, async_app):
    """aiohttp client fixture to ease testing

    source: https://docs.aiohttp.org/en/stable/testing.html
    """
    loop.set_debug(True)
    return loop.run_until_complete(aiohttp_client(async_app))


async def test_api_async_endpoint(cli, async_app):
    res = await cli.post(
        "/path/to/identity",
        headers=[("Content-Type", "application/json"), ("Accept", "application/json")],
        data=json_dumps({"data": "toto"}),
    )
    assert res.status == 200
    assert res.content_type == "application/json"
    assert await res.read() == json_dumps("toto").encode()


async def test_api_async_nego_default_msgpack(cli):
    res = await cli.post(
        "/path/to/identity",
        headers=[("Content-Type", "application/json")],
        data=json_dumps({"data": "toto"}),
    )
    assert res.status == 200
    assert res.content_type == "application/x-msgpack"
    assert await res.read() == msgpack_dumps("toto")


async def test_api_async_nego_default(cli):
    res = await cli.post(
        "/path/to/identity",
        headers=[
            ("Content-Type", "application/json"),
            ("Accept", "application/x-msgpack"),
        ],
        data=json_dumps({"data": "toto"}),
    )
    assert res.status == 200
    assert res.content_type == "application/x-msgpack"
    assert await res.read() == msgpack_dumps("toto")


async def test_api_async_nego_accept(cli):
    res = await cli.post(
        "/path/to/identity",
        headers=[
            ("Accept", "application/x-msgpack"),
            ("Content-Type", "application/x-msgpack"),
        ],
        data=msgpack_dumps({"data": "toto"}),
    )
    assert res.status == 200
    assert res.content_type == "application/x-msgpack"
    assert await res.read() == msgpack_dumps("toto")


async def test_api_async_rpc_server(cli):
    res = await cli.get(
        "/test_endpoint_url",
        headers=[
            ("Content-Type", "application/x-msgpack"),
            ("Accept", "application/x-msgpack"),
        ],
        data=msgpack_dumps({"test_data": "spam"}),
    )

    assert res.status == 200
    assert res.content_type == "application/x-msgpack"
    assert await res.read() == msgpack_dumps("egg")


async def test_api_async_rpc_server_extra_serializers(cli):
    res = await cli.post(
        "/serializer_test",
        headers=[
            ("Content-Type", "application/x-msgpack"),
            ("Accept", "application/x-msgpack"),
        ],
        data=(
            b"\x81\xa4data\x92\xa3foo\x82\xc4\x07swhtype\xa9extratype"
            b"\xc4\x01d\x92\xa3bar\xc4\x03baz"
        ),
    )

    assert res.status == 200
    assert res.content_type == "application/x-msgpack"
    assert await res.read() == (
        b"\x82\xc4\x07swhtype\xa9extratype\xc4\x01d\x92\x81\xa4spam\xa3egg\xa3qux"
    )



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import inspect


def remove_kwargs(names):
    def decorator(f):
        sig = inspect.signature(f)
        params = sig.parameters
        params = [param for param in params.values() if param.name not in names]
        sig = sig.replace(parameters=params)
        f.__signature__ = sig
        return f

    return decorator


def apply_options(cursor, options):
    """Applies the given postgresql client options to the given cursor.

    Returns a dictionary with the old values if they changed."""
    old_options = {}
    for option, value in options.items():
        cursor.execute("SHOW %s" % option)
        old_value = cursor.fetchall()[0][0]
        if old_value != value:
            cursor.execute("SET LOCAL %s TO %%s" % option, (value,))
            old_options[option] = old_value
    return old_options


def db_transaction(**client_options):
    """decorator to execute Backend methods within DB transactions

    The decorated method must accept a `cur` and `db` keyword argument

    Client options are passed as `set` options to the postgresql server
    """

    def decorator(meth, __client_options=client_options):
        if inspect.isgeneratorfunction(meth):
            raise ValueError("Use db_transaction_generator for generator functions.")

        @remove_kwargs(["cur", "db"])
        @functools.wraps(meth)
        def _meth(self, *args, **kwargs):
            if "cur" in kwargs and kwargs["cur"]:
                cur = kwargs["cur"]
                old_options = apply_options(cur, __client_options)
                ret = meth(self, *args, **kwargs)
                apply_options(cur, old_options)
                return ret
            else:
                db = self.get_db()
                try:
                    with db.transaction() as cur:
                        apply_options(cur, __client_options)
                        return meth(self, *args, db=db, cur=cur, **kwargs)
                finally:
                    self.put_db(db)

        return _meth

    return decorator


def db_transaction_generator(**client_options):
    """decorator to execute Backend methods within DB transactions, while
    returning a generator

    The decorated method must accept a `cur` and `db` keyword argument

    Client options are passed as `set` options to the postgresql server
    """

    def decorator(meth, __client_options=client_options):
        if not inspect.isgeneratorfunction(meth):
            raise ValueError("Use db_transaction for non-generator functions.")

        @remove_kwargs(["cur", "db"])
        @functools.wraps(meth)
        def _meth(self, *args, **kwargs):
            if "cur" in kwargs and kwargs["cur"]:
                cur = kwargs["cur"]
                old_options = apply_options(cur, __client_options)
                yield from meth(self, *args, **kwargs)
                apply_options(cur, old_options)
            else:
                db = self.get_db()
                try:
                    with db.transaction() as cur:
                        apply_options(cur, __client_options)
                        yield from meth(self, *args, db=db, cur=cur, **kwargs)
                finally:
                    self.put_db(db)

        return _meth

    return decorator



import os

from hypothesis import HealthCheck

os.environ["LC_ALL"] = "C.UTF-8"

# we use getattr here to keep mypy happy regardless hypothesis version
function_scoped_fixture_check = (
    [getattr(HealthCheck, "function_scoped_fixture")]
    if hasattr(HealthCheck, "function_scoped_fixture")
    else []
)



#!/usr/bin/env python3
# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from os import environ, path
from typing import Collection, Dict, Optional, Tuple
import warnings

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

warnings.filterwarnings("ignore")  # noqa prevent psycopg from telling us sh*t


logger = logging.getLogger(__name__)


@swh_cli_group.group(name="db", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Configuration file.",
)
@click.pass_context
def db(ctx, config_file):
    """Software Heritage database generic tools."""
    from swh.core.config import read as config_read

    ctx.ensure_object(dict)
    if config_file is None:
        config_file = environ.get("SWH_CONFIG_FILENAME")
    cfg = config_read(config_file)
    ctx.obj["config"] = cfg


@db.command(name="create", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name.",
    default="softwareheritage-dev",
    show_default=True,
)
@click.option(
    "--template",
    "-T",
    help="Template database from which to build this database.",
    default="template1",
    show_default=True,
)
def db_create(module, dbname, template):
    """Create a database for the Software Heritage <module>.

    and potentially execute superuser-level initialization steps.

    Example::

        swh db create -d swh-test storage

    If you want to specify non-default postgresql connection parameters, please
    provide them using standard environment variables or by the mean of a
    properly crafted libpq connection URI. See psql(1) man page (section
    ENVIRONMENTS) for details.

    Note: this command requires a postgresql connection with superuser permissions.

    Example::

        PGPORT=5434 swh db create indexer
        swh db create -d postgresql://superuser:passwd@pghost:5433/swh-storage storage

    """

    logger.debug("db_create %s dn_name=%s", module, dbname)
    create_database_for_package(module, dbname, template)


@db.command(name="init-admin", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name.",
    default="softwareheritage-dev",
    show_default=True,
)
def db_init_admin(module: str, dbname: str) -> None:
    """Execute superuser-level initialization steps (e.g pg extensions, admin functions,
    ...)

    Example::

        PGPASSWORD=... swh db init-admin -d swh-test scheduler

    If you want to specify non-default postgresql connection parameters, please
    provide them using standard environment variables or by the mean of a
    properly crafted libpq connection URI. See psql(1) man page (section
    ENVIRONMENTS) for details.

    Note: this command requires a postgresql connection with superuser permissions (e.g
    postgres, swh-admin, ...)

    Example::

        PGPORT=5434 swh db init-admin scheduler
        swh db init-admin -d postgresql://superuser:passwd@pghost:5433/swh-scheduler \
          scheduler

    """
    logger.debug("db_init_admin %s dbname=%s", module, dbname)
    init_admin_extensions(module, dbname)


@db.command(name="init", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name.",
    default="softwareheritage-dev",
    show_default=True,
)
@click.option(
    "--flavor", help="Database flavor.", default=None,
)
def db_init(module, dbname, flavor):
    """Initialize a database for the Software Heritage <module>.

    Example::

        swh db init -d swh-test storage

    If you want to specify non-default postgresql connection parameters,
    please provide them using standard environment variables.
    See psql(1) man page (section ENVIRONMENTS) for details.

    Examples::

        PGPORT=5434 swh db init indexer
        swh db init -d postgresql://user:passwd@pghost:5433/swh-storage storage
        swh db init --flavor read_replica -d swh-storage storage

    """

    logger.debug("db_init %s flavor=%s dbname=%s", module, flavor, dbname)

    initialized, dbversion, dbflavor = populate_database_for_package(
        module, dbname, flavor
    )

    # TODO: Ideally migrate the version from db_version to the latest
    # db version

    click.secho(
        "DONE database for {} {}{} at version {}".format(
            module,
            "initialized" if initialized else "exists",
            f" (flavor {dbflavor})" if dbflavor is not None else "",
            dbversion,
        ),
        fg="green",
        bold=True,
    )

    if flavor is not None and dbflavor != flavor:
        click.secho(
            f"WARNING requested flavor '{flavor}' != recorded flavor '{dbflavor}'",
            fg="red",
            bold=True,
        )


def get_sql_for_package(modname):
    import glob
    from importlib import import_module

    from swh.core.utils import numfile_sortkey as sortkey

    if not modname.startswith("swh."):
        modname = "swh.{}".format(modname)
    try:
        m = import_module(modname)
    except ImportError:
        raise click.BadParameter("Unable to load module {}".format(modname))

    sqldir = path.join(path.dirname(m.__file__), "sql")
    if not path.isdir(sqldir):
        raise click.BadParameter(
            "Module {} does not provide a db schema " "(no sql/ dir)".format(modname)
        )
    return sorted(glob.glob(path.join(sqldir, "*.sql")), key=sortkey)


def populate_database_for_package(
    modname: str, conninfo: str, flavor: Optional[str] = None
) -> Tuple[bool, int, Optional[str]]:
    """Populate the database, pointed at with ``conninfo``,
    using the SQL files found in the package ``modname``.

    Args:
      modname: Name of the module of which we're loading the files
      conninfo: connection info string for the SQL database
      flavor: the module-specific flavor which we want to initialize the database under

    Returns:
      Tuple with three elements: whether the database has been initialized; the current
      version of the database; if it exists, the flavor of the database.
    """
    from swh.core.db.db_utils import swh_db_flavor, swh_db_version

    current_version = swh_db_version(conninfo)
    if current_version is not None:
        dbflavor = swh_db_flavor(conninfo)
        return False, current_version, dbflavor

    sqlfiles = get_sql_for_package(modname)
    sqlfiles = [fname for fname in sqlfiles if "-superuser-" not in fname]
    execute_sqlfiles(sqlfiles, conninfo, flavor)

    current_version = swh_db_version(conninfo)
    assert current_version is not None
    dbflavor = swh_db_flavor(conninfo)
    return True, current_version, dbflavor


def parse_dsn_or_dbname(dsn_or_dbname: str) -> Dict[str, str]:
    """Parse a psycopg2 dsn, falling back to supporting plain database names as well"""
    import psycopg2
    from psycopg2.extensions import parse_dsn as _parse_dsn

    try:
        return _parse_dsn(dsn_or_dbname)
    except psycopg2.ProgrammingError:
        # psycopg2 failed to parse the DSN; it's probably a database name,
        # handle it as such
        return _parse_dsn(f"dbname={dsn_or_dbname}")


def init_admin_extensions(modname: str, conninfo: str) -> None:
    """The remaining initialization process -- running -superuser- SQL files -- is done
    using the given conninfo, thus connecting to the newly created database

    """
    sqlfiles = get_sql_for_package(modname)
    sqlfiles = [fname for fname in sqlfiles if "-superuser-" in fname]
    execute_sqlfiles(sqlfiles, conninfo)


def create_database_for_package(
    modname: str, conninfo: str, template: str = "template1"
):
    """Create the database pointed at with ``conninfo``, and initialize it using
    -superuser- SQL files found in the package ``modname``.

    Args:
      modname: Name of the module of which we're loading the files
      conninfo: connection info string or plain database name for the SQL database
      template: the name of the database to connect to and use as template to create
        the new database

    """
    import subprocess

    from psycopg2.extensions import make_dsn

    # Use the given conninfo string, but with dbname replaced by the template dbname
    # for the database creation step
    creation_dsn = parse_dsn_or_dbname(conninfo)
    dbname = creation_dsn["dbname"]
    creation_dsn["dbname"] = template
    logger.debug("db_create dbname=%s (from %s)", dbname, template)
    subprocess.check_call(
        [
            "psql",
            "--quiet",
            "--no-psqlrc",
            "-v",
            "ON_ERROR_STOP=1",
            "-d",
            make_dsn(**creation_dsn),
            "-c",
            f'CREATE DATABASE "{dbname}"',
        ]
    )
    init_admin_extensions(modname, conninfo)


def execute_sqlfiles(
    sqlfiles: Collection[str], conninfo: str, flavor: Optional[str] = None
):
    """Execute a list of SQL files on the database pointed at with ``conninfo``.

    Args:
      sqlfiles: List of SQL files to execute
      conninfo: connection info string for the SQL database
      flavor: the database flavor to initialize
    """
    import subprocess

    psql_command = [
        "psql",
        "--quiet",
        "--no-psqlrc",
        "-v",
        "ON_ERROR_STOP=1",
        "-d",
        conninfo,
    ]

    flavor_set = False
    for sqlfile in sqlfiles:
        logger.debug(f"execute SQL file {sqlfile} dbname={conninfo}")
        subprocess.check_call(psql_command + ["-f", sqlfile])

        if flavor is not None and not flavor_set and sqlfile.endswith("-flavor.sql"):
            logger.debug("Setting database flavor %s", flavor)
            query = f"insert into dbflavor (flavor) values ('{flavor}')"
            subprocess.check_call(psql_command + ["-c", query])
            flavor_set = True

    if flavor is not None and not flavor_set:
        logger.warn(
            "Asked for flavor %s, but module does not support database flavors", flavor,
        )
